package watcher

import (
	"context"
	"fmt"
	"github.com/golang/glog"
	"gorm.io/gorm"
	"gorm.io/gorm/clause"
	"gorm.io/gorm/schema"
	"haitao_watcher/pkg/model"
	"haitao_watcher/pkg/options"
	pool "haitao_watcher/pkg/pools"
	"sync"
	"time"
)

type Controller struct {
	lock sync.RWMutex
	m    map[string]Watcher
	db   *gorm.DB
	pool *pool.ProxyPool
	ctx  context.Context
	ch   chan<- model.PushMsg
}

func NewController(ctx context.Context, db *gorm.DB, pool *pool.ProxyPool, ch chan<- model.PushMsg) *Controller {
	ctl := &Controller{
		m:    make(map[string]Watcher),
		db:   db,
		pool: pool,
		ctx:  ctx,
		ch:   ch,
	}
	ctl.migrateTables()
	go func() {
		if err := ctl.initWatchers(); err != nil {
			glog.Errorf("watcher init :%v", err)
		}
	}()
	return ctl
}

func (c *Controller) migrateTables() {
	tables := []schema.Tabler{
		&model.Product{},
	}
	for _, table := range tables {
		if err := c.db.AutoMigrate(table); err != nil {
			glog.Fatalf("failed to migrate table %s: %v", table.TableName(), err)
		}
	}
}

func (c *Controller) initWatchers() error {
	var list []model.Product
	if err := c.db.Find(&list, "orderable = ?", false).Error; err != nil {
		return err
	}
	for _, p := range list {
		if err := c.RunWatcher(&options.CoachOutletOption{
			Pid:       p.Pid,
			Remark:    p.Remark,
			Interval:  time.Minute * 5,
			PusherIds: p.PusherIds,
		}); err != nil {
			glog.Errorf("run watcher err: %v", err)
		}
	}
	return nil
}

func (c *Controller) RunWatcher(opt *options.CoachOutletOption) error {
	uid := opt.Uid()
	err := c.db.Clauses(clause.OnConflict{
		Columns:   []clause.Column{{Name: "uid"}},
		DoUpdates: clause.AssignmentColumns([]string{"pid", "watch", "remark", "orderable", "pusher_ids", "deleted_at"}),
	}).Create(&model.Product{Uid: uid, Pid: opt.Pid, Watch: true, Remark: opt.Remark, Orderable: false, PusherIds: opt.PusherIds}).Error

	if err != nil {
		return err
	}

	if water, ok := c.m[uid]; ok {
		water.Restart()
		return nil
	}

	watcher := NewCoachWatcher(c.ctx, c.pool, opt, c.db).SetOnOrderableMsgChannel(c.ch)
	go watcher.Watch()

	c.lock.Lock()
	c.m[watcher.Uid()] = watcher
	c.lock.Unlock()
	return nil
}

type ListWatcherInfoRequest struct {
	Website   model.WebsiteType `query:"website,omitempty"` //是什么网站
	Watch     *bool             `query:"watch,omitempty"`
	Orderable *bool             `query:"orderable,omitempty"`
	Keyword   string            `query:"keyword,omitempty"`
	Page      int               `query:"page"`
	Size      int               `query:"size"`
}

func (c *Controller) List(req ListWatcherInfoRequest) (resp model.ListResponse[model.Product], err error) {
	tx := c.db
	if req.Keyword != "" {
		tx = tx.Where("name LIKE ? or remark LIKE ?", fmt.Sprintf("%%%s%%", req.Keyword), fmt.Sprintf("%%%s%%", req.Keyword))
	}
	if !req.Website.IsZero() {
		tx = tx.Where("website = ?", req.Website)
	}
	if req.Watch != nil {
		tx = tx.Where("watch = ?", req.Watch)
	}
	if req.Orderable != nil {
		tx = tx.Where("orderable = ?", req.Orderable)
	}
	if err = tx.Model(&model.Product{}).Count(&resp.Total).Error; err != nil {
		return resp, fmt.Errorf("查询总数失败：%v", err)
	}
	if resp.Total == 0 {
		resp.List = make([]model.Product, 0)
		return
	}
	// 查询列表
	if req.Page < 1 {
		req.Page = 1
	}
	if req.Size < 1 {
		req.Size = 10
	}
	offset := (req.Page - 1) * req.Size

	if err = tx.Order("created_at desc").Limit(req.Size).Offset(offset).
		Find(&resp.List).Error; err != nil {
		return resp, fmt.Errorf("查询列表失败：%v", err)
	}
	return
}

func (c *Controller) Delete(uid string) error {
	c.lock.Lock()
	if watcher, ok := c.m[uid]; ok {
		watcher.Cancel()
		delete(c.m, uid)
	}
	c.lock.Unlock()
	return c.db.Delete(&model.Product{Uid: uid}, "uid = ?", uid).Error
}

func (c *Controller) Stop(uid string) error {
	c.lock.RLock()
	if watcher, ok := c.m[uid]; ok {
		watcher.Cancel()
	}
	c.lock.RUnlock()
	return c.db.Model(&model.Product{}).Where("uid = ?", uid).UpdateColumn("watch", false).Error
}

func (c *Controller) Start(uid string) error {
	c.lock.RLock()
	if watcher, ok := c.m[uid]; ok {
		watcher.Restart()
		c.lock.RUnlock()
		return nil
	}
	c.lock.RUnlock()

	var product model.Product
	if err := c.db.Find(&product, "uid = ?", uid).Error; err != nil {
		return err
	}
	return c.RunWatcher(&options.CoachOutletOption{
		Pid:       product.Pid,
		Remark:    product.Remark,
		Interval:  time.Minute * 5,
		PusherIds: product.PusherIds,
	})
}

func (c *Controller) SetPushers(uid string, pushers []uint) (err error) {
	c.lock.RLock()
	if watcher, ok := c.m[uid]; ok {
		cfg := watcher.Option()
		cfg.PusherIds = pushers
		watcher.SetOption(cfg)
	}
	c.lock.RUnlock()
	return c.db.Model(&model.Product{}).Where("uid = ?", uid).UpdateColumn("pusher_ids", pushers).Error
}
